import { CaseReducer, createSlice, PayloadAction } from '@reduxjs/toolkit';
import type { RootState } from './store';
import { DatabaseStatus, SaveStateI } from '../broker/wsState';
import { getParam, URLParams } from '../api/url';

/** Message format of the error message from the backend */
export type ErrorMessage = {
  errorStatus: any;
  queryID: any;
};

/** Cache type */
export type SessionCacheI = {
  currentSaveState?: string; // id of the current save state
  saveStates: Record<string, SaveStateI>;
  testedSaveState: Record<string, DatabaseStatus>;
};

// Define the initial state using that type
export const initialState: SessionCacheI = {
  currentSaveState: undefined,
  saveStates: {},
  testedSaveState: {},
};

//@ts-ignore(2589)
export const sessionSlice = createSlice({
  name: 'session',
  initialState: initialState,
  reducers: {
    selectSaveState: (state: SessionCacheI, action: PayloadAction<string | undefined>) => {
      if (action.payload && action.payload in state.saveStates) {
        state.currentSaveState = action.payload;
      } else {
        state.currentSaveState = undefined;
      }
    },
    updateSelectedSaveState: (state: SessionCacheI, action: PayloadAction<SaveStateI>) => {
      if (state.currentSaveState === action.payload.id && state.currentSaveState) state.saveStates[state.currentSaveState] = action.payload;
    },
    updateSaveStateList: (state: SessionCacheI, action: PayloadAction<SaveStateI[]>) => {
      // Does NOT clear the states, just adds in new data
      let newState: Record<string, SaveStateI> = {};
      action.payload.forEach((ss: SaveStateI) => {
        newState[ss.id] = ss;
      });
      state.saveStates = { ...state.saveStates, ...newState };
      const paramSaveState = getParam(URLParams.saveState);
      if (!state.currentSaveState) {
        if (paramSaveState && paramSaveState in state.saveStates) {
          state.currentSaveState = paramSaveState;
        } else if (Object.keys(state.saveStates).length > 0) {
          state.currentSaveState = Object.keys(state.saveStates)[0];
        } else state.currentSaveState = undefined;
      }
    },
    setSaveStateList: (state: SessionCacheI, action: PayloadAction<SaveStateI[]>) => {
      // Clears the states and puts in new data
      let newState: Record<string, SaveStateI> = {};
      action.payload.forEach((ss) => {
        newState[ss.id] = ss;
      });
      state.saveStates = newState;
      const paramSaveState = getParam(URLParams.saveState);
      if (!state.currentSaveState) {
        if (paramSaveState && paramSaveState in state.saveStates) {
          state.currentSaveState = paramSaveState;
        } else if (Object.keys(state.saveStates).length > 0) {
          state.currentSaveState = Object.keys(state.saveStates)[0];
        } else state.currentSaveState = undefined;
      }
    },
    addSaveState: (state: SessionCacheI, action: PayloadAction<SaveStateI>) => {
      if (state.saveStates === undefined) state.saveStates = {};
      state.saveStates[action.payload.id] = action.payload;
      state.currentSaveState = action.payload.id;
    },
    deleteSaveState: (state: SessionCacheI, action: PayloadAction<string>) => {
      delete state.saveStates[action.payload];
      if (state.currentSaveState === action.payload) {
        if (Object.keys(state.saveStates).length > 0) state.currentSaveState = Object.keys(state.saveStates)[0];
        else state.currentSaveState = undefined;
      }
    },
    testedSaveState: (state: SessionCacheI, action: PayloadAction<string>) => {
      state.testedSaveState = { ...state.testedSaveState, [action.payload]: DatabaseStatus.tested };
    },
  },
});

export const {
  selectSaveState,
  deleteSaveState,
  updateSaveStateList,
  setSaveStateList,
  addSaveState,
  testedSaveState,
  updateSelectedSaveState,
} = sessionSlice.actions;

// Other code such as selectors can use the imported `RootState` type
export const sessionCacheState = (state: RootState) => state.sessionCache;

export default sessionSlice.reducer;
